#include <M5StickC.h>
#include <WiFi.h>
#include <SoftwareSerial.h>
#include <HTTPClient.h>
#include "Adafruit_BME680.h"
#include "Adafruit_TSL2591.h"
#include "MHZ19.h"

#define I2C_SDA         (0)
#define I2C_SCL         (26)
#define MHZ19_UART_TX   (33)
#define MHZ19_UART_RX   (32)
#define MHZ19_BAUDRATE  (9600)
#define PREHEATING_TIME_MS      (3 * 60 * 1000) // Co2センサ予熱時間（３分）
#define TRANSMIT_INTERVAL_MS    (3 * 60 * 1000)
#define BTN_HOLD_TIME_FOR_CALIB (4000) //[ms]
#define NTP_SERVER1     "ntp.nict.jp"
#define NTP_SERVER2     "ntp.jst.mfeed.ad.jp"
#define NUM_SENSORS     (7)
#define GAS_HOST        "https://script.google.com/macros/s/*****/exec"
#define VOLTAGE_HIGH    (4.8) // 電源ONとする電圧
#define VOLTAGE_LOW     (3.6) // 電源OFFとする電圧
#define BME680_TEMP_OFFSET      (-1.4)

const char *ssid = "<モバイルルータのSSID>";
const char *password = "<モバイルルータのパスワード>";
const char *sensors[NUM_SENSORS] = {"Temp", "Humid", "Press", "VOCs", "Light", "IR-Light", "Co2"};

/* WDT Timeout time (us) */
const unsigned long wdtTimeout = 60 * 1000 * 1000;
hw_timer_t *wdt = NULL;

bool isPreHeated = false;

WiFiClient client;
Adafruit_BME680 bme;
Adafruit_TSL2591 tsl;
MHZ19 mhz;
HTTPClient http;
float lastUsbVoltage;
unsigned long lastTimestamp;

/*
 * Reset process by interrupt(WDT)
 */
void IRAM_ATTR resetModule() {
  esp_restart();
}

/*
 * NTPを使ってローカル時間を設定
 */
void initLocalTime() {
  configTzTime("JST-9", NTP_SERVER1, NTP_SERVER2);

  struct tm timeInfo;
  if (getLocalTime(&timeInfo)) {
    M5.Lcd.printf("%d/%02d/%02d\n",
                  timeInfo.tm_year + 1900,
                  timeInfo.tm_mon + 1,
                  timeInfo.tm_mday);
    M5.Lcd.printf("%02d:%02d:%02d\n",
                  timeInfo.tm_hour,
                  timeInfo.tm_min,
                  timeInfo.tm_sec);
                  
  } else {
    /* 時間取得できない場合はリセット */
    esp_restart();
  }
  delay(500);
}

/*
 * Setup
 */
void setup() {
  /* WDT設定 */
  wdt = timerBegin(0, 80, true); // Timer Div : 80(=1us)
  timerAttachInterrupt(wdt, &resetModule, true);
  timerAlarmWrite(wdt, wdtTimeout, false);
  timerAlarmEnable(wdt);

  M5.begin();
  M5.Lcd.fillScreen(BLACK);
  M5.Lcd.setRotation(2);
  M5.Lcd.setCursor(0, 0, 1);

  Wire.begin(I2C_SDA, I2C_SCL);

  /* BME680設定 */
  while(!bme.begin()) {
    delay(500);
  }
  bme.setTemperatureOversampling(BME680_OS_4X);
  bme.setHumidityOversampling(BME680_OS_2X);
  bme.setPressureOversampling(BME680_OS_4X);
  bme.setIIRFilterSize(BME680_FILTER_SIZE_1);
  bme.setGasHeater(320, 150);
  
  /* TSL2591設定*/
  tsl.setGain(TSL2591_GAIN_LOW);
  tsl.setTiming(TSL2591_INTEGRATIONTIME_100MS);

  /* MHZ19B設定 */
  //ss.begin(MHZ19_BAUDRATE);
  Serial2.begin(MHZ19_BAUDRATE, SERIAL_8N1, MHZ19_UART_RX, MHZ19_UART_TX);
  mhz.begin(Serial2);
  mhz.autoCalibration(false);

  /* WiFi接続しNTPで時刻合わせ */
  M5.Lcd.println("WiFi");
  M5.Lcd.println("Connecting...");
  
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
  }
  M5.Lcd.println(WiFi.localIP());

  initLocalTime();

  lastUsbVoltage = M5.Axp.GetVBusVoltage();
  lastTimestamp = millis();
}

/*
 * 全てのセンサデータを読み取り表示，JSON文字列に変換する．
 * @return センサデータのJSON文字列
 */
String readSensor() {
  /* 各センサからデータを読み取る */
  float senData[NUM_SENSORS];
  bme.performReading();
  senData[0] = bme.temperature + BME680_TEMP_OFFSET;
  senData[1] = bme.humidity;
  senData[2] = bme.pressure / 100.0;  // hPa
  senData[3] = bme.gas_resistance / 1000.0;  // Kohm

  uint32_t lum = tsl.getFullLuminosity();
  senData[4] = (float)(lum & 0xFFFF);
  senData[5] = (float)(lum >> 16);
  
  senData[6] = (float)mhz.getCO2(false);

  /* ディスプレイ表示 */
  char date_str[11], time_str[9];
  struct tm timeInfo;
  getLocalTime(&timeInfo);
  sprintf(date_str, "%d/%02d/%02d",
          timeInfo.tm_year+1900,
          timeInfo.tm_mon+1,
          timeInfo.tm_mday);
  sprintf(time_str, "%02d:%02d:%02d",
          timeInfo.tm_hour,
          timeInfo.tm_min,
          timeInfo.tm_sec);

  M5.Lcd.fillScreen(BLACK);
  M5.Lcd.setCursor(0, 0, 1);
  M5.Lcd.println(date_str);
  M5.Lcd.println(time_str);
  for (int i = 0; i < NUM_SENSORS; i++) {
    M5.Lcd.println(sensors[i]);
    M5.Lcd.printf("      %.1f\n", senData[i]);
  }
  
  /* JSON文字列へ変換 */
  String json = "{\"Type\":\"Sensors\",";
  json += "\"Data\":{";
  for (int i = 0; i < NUM_SENSORS; i++) {
    if (i > 0) json += ",";
    char buf[10];
    dtostrf(senData[i], 4, 2, buf);
    json += "\"" + String(sensors[i]) + "\":";
    json += String(buf);
  }
  json += "},\"Timestamp\":";
  json += "\"" + String(date_str) + " " + String(time_str) + "\"";
  json += "}";
  return json;
}

/*
 * JSONデータをGASへ送信する
 * @return true=通信成功 false=通信失敗
 */
bool sendJson(String json) {
  http.begin(GAS_HOST);
  http.addHeader("Content-Type", "application/json");
  const char* buf = json.c_str();
  int res_code = http.POST((uint8_t*)buf, strlen(buf));
  http.end();
  return (res_code > 0);
}

/*
 * main loop
 */
void loop() {
  timerWrite(wdt, 0); // WDT Clear

  M5.update();

  /* 電源ケーブルが抜けたら即時通知 */
  float currentVoltage = M5.Axp.GetVBusVoltage();
  if (lastUsbVoltage > VOLTAGE_HIGH && currentVoltage < VOLTAGE_LOW) {
    String json = "{\"Type\":\"Emergency\",\"Message\":\"UsbVoltageLow\"}";
    sendJson(json);
  }
  lastUsbVoltage = currentVoltage;
  
  /* Aボタン長押しでCO2センサのゼロキャリブレーション */
  if (M5.BtnA.wasReleasefor(BTN_HOLD_TIME_FOR_CALIB)) {
    M5.Lcd.fillScreen(BLACK);
    M5.Lcd.setCursor(0, 0, 1);
    M5.Lcd.println("CO2");
    M5.Lcd.println("Zero");
    M5.Lcd.println("Calibration");
    mhz.calibrateZero();
  }

  /* 定期的にセンサを読み取り送信 */  
  if ((millis() - lastTimestamp) >= TRANSMIT_INTERVAL_MS) {
    String json = readSensor();

    if (!isPreHeated) {
      /* センサ予熱中のデータは送信しない */
      if (millis() > PREHEATING_TIME_MS) {
        isPreHeated = true;
      }
    } else {
      while(!sendJson(json)) {
        delay(1000);
      }
    }
    
    lastTimestamp = millis();
  }
}
